import pickle
from datetime import datetime
from typing import Union, List

from portfolio_manager.exceptions import (InsufficientFunds, BackDatingError,
                                          InsufficientData)


class InvestmentPortfolio:
    def __init__(self,
                 name: str = None,
                 total_deposited: Union[int, float] = None,
                 current_portfolio_value: Union[int, float] = None,
                 portfolio_history: List[dict] = None):
        """
        Represents an investment portfolio. Funds can be deposited/withdrawn, and the
        current value of the assets in the portfolio can be updated over time. Historical
        transactions and value changes are saved.

        Parameters
        ----------
        name : str
            The name of the portfolio. Defaults to 'portfolio_{today's date}'.
        total_deposited : Union[int, float]
            The total amount that has been invested into the portfolio. i.e. the sum of
            each individual deposit.
        current_portfolio_value : Union[int, float]
            The current total value of all the assets within the portfolio.
        portfolio_history : List[dict]
            Stores historical portfolio data by saving a snapshot of the portfolio each
            time it is updated. Each snapshot is a dictionary with keys 'date',
            'total_deposited', 'current_portfolio_value' and 'transaction_type'.
        """
        date_today = datetime.now().strftime('%d%m%Y')
        self.name = name or f'portfolio_{date_today}'
        self.total_deposited = total_deposited or 0
        self.current_portfolio_value = current_portfolio_value or 0
        self.portfolio_history = portfolio_history or []

        # Used internally for catching back-dating errors
        self.latest_transaction_date = None

    def deposit(self,
                deposit_amount: Union[int, float],
                portfolio_value_before_deposit: Union[int, float] = None,
                date: datetime = None):
        """
        Deposit funds into the portfolio.

        Parameters
        ----------
        deposit_amount : Union[int, float]
            The amount of money being deposited into the portfolio.
        portfolio_value_before_deposit : Union[int, float]
            The total value of the portfolio before making the deposit.
        date : datetime
            When the deposit was made. Defaults to now.
        """
        date = date or datetime.now()
        self._backdate_error_check(date)

        # Update the portfolio value from before the deposit, if this value is provided
        if portfolio_value_before_deposit:
            self.update_portfolio_value(portfolio_value_before_deposit, date)

        # Update the total amount deposited and current portfolio value
        self.total_deposited += deposit_amount
        self.current_portfolio_value += deposit_amount

        # Update portfolio_history
        self._update_portfolio_history(date, 'deposit')

    def withdraw(self,
                 withdrawal_amount: Union[int, float],
                 portfolio_value_before_withdrawal: Union[int, float] = None,
                 date: datetime = None):
        """
        Deposit additional funds into the portfolio.

        Parameters
        ----------
        withdrawal_amount : Union[int, float]
            The amount of money being withdrawn from the portfolio.
        portfolio_value_before_withdrawal : Union[int, float]
            The total value of the portfolio before making the withdrawal.
        date : datetime
            When the withdrawal was made. Defaults to now.
        """
        date = date or datetime.now()
        self._backdate_error_check(date)

        # Update the portfolio value from before the deposit, if this value is provided
        if portfolio_value_before_withdrawal:
            self.update_portfolio_value(portfolio_value_before_withdrawal, date)

        # If trying to withdraw more than the portfolio value, raise an error
        if withdrawal_amount > self.current_portfolio_value:
            raise InsufficientFunds(f'Cannot withdraw more than the portfolio value: '
                                    f'{self.current_portfolio_value}')

        # Update the total amount deposited and the current portfolio value
        self.total_deposited -= withdrawal_amount
        self.current_portfolio_value -= withdrawal_amount

        # Update portfolio_history
        self._update_portfolio_history(date, 'withdrawal')

    def update_portfolio_value(self, current_portfolio_value: Union[int, float],
                               date: datetime = None):
        """
        Update the current value of all assets in the portfolio.

        Parameters
        ----------
        current_portfolio_value : Union[int, float]
            The current total value of all assets in the portfolio.
        date : datetime
            When this valuation was calculated. Defaults to now.
        """
        if len(self.portfolio_history) == 0:
            raise InsufficientData("First transaction can't be a value update; make a "
                                   "deposit!")

        date = date or datetime.now()
        self._backdate_error_check(date)

        # Update the current total value of the assets in the portfolio
        self.current_portfolio_value = current_portfolio_value

        # Update portfolio_history
        self._update_portfolio_history(date, 'update_portfolio_value')

    def _update_portfolio_history(self, date: datetime, transaction_type: str):
        """
        Add a snapshot of the current portfolio to portfolio_history.

        Parameters
        ----------
        date : datetime
            The date at which the portfolio snapshot was taken.
        transaction_type : str
            The type of transaction that was made before taking the portfolio snapshot.
        """
        # Update portfolio_history
        new_entry = {
            'date': date,
            'total_deposited': self.total_deposited,
            'current_portfolio_value': self.current_portfolio_value,
            'transaction_type': transaction_type
        }
        portfolio_history = self.portfolio_history
        portfolio_history.append(new_entry)

        # Ensure the portfolio history is stored in order of ascending transaction date
        sorted_portfolio_history = sorted(portfolio_history, key=lambda x: x['date'])
        self.portfolio_history = sorted_portfolio_history
        self.latest_transaction_date = self.portfolio_history[-1]['date']

    def _backdate_error_check(self, date):
        """ Ensure that the transaction being made isn't being incorrectly back-dated """
        if self.latest_transaction_date is not None:
            if date <= self.latest_transaction_date:
                raise BackDatingError(
                    f'Attempted transaction: {date}. Latest portfolio transaction: '
                    f'{self.latest_transaction_date}.')

    def save_portfolio(self, directory: str = None):
        """
        Save the state of the current portfolio. The portfolio object is pickled and
        saved within the specified directory under the name '{self.name}.pkl'.

        Parameters
        ----------
        directory : str
            The directory that the pickled object should be saved in. Defaults to the
            current working directory.
        """
        path = f'{directory}/{self.name}.pkl' if directory else f'{self.name}.pkl'

        # Save the portfolio, overwriting any existing files at that path.
        with open(path, 'wb') as handle:
            pickle.dump(self, handle, pickle.HIGHEST_PROTOCOL)


def load_portfolio(name: str, directory: str = None) -> InvestmentPortfolio:
    """
    Load a previously pickled and saved portfolio from the specified path.

    Parameters
    ----------
    name : str
        The name of the portfolio being loaded. i.e. the 'name' attribute of the saved
        portfolio object.
    directory : str
        The directory containing the saved portfolio. By default, look in the current
        working directory.

    Returns
    -------
    InvestmentPortfolio : The instantiated portfolio object.
    """
    # Add the .pkl file extension if it isn't already there
    if '.' not in name:
        name += '.pkl'

    # Load and return the portfolio
    path = f'{directory}/{name}' if directory else name
    with open(path, 'rb') as handle:
        portfolio = pickle.load(handle)

    return portfolio



class PortfolioError(Exception):
    """ Base class for handling exceptions related to investment portfolios. """
    pass


class BackDatingError(PortfolioError):
    """
    Attempting to back-date a transaction too far into the past. Transactions can not be
    back-dated to further back than the most recent transaction. This is to prevent
    issues with inserting transactions which make future transactions invalid.

    e.g. If the most recent transaction was made on 1st Jan 2021 12:00pm, then a
        transaction can be back-dated to 1st Jan 2021 12:01pm, but no earlier.
    """
    pass


class InsufficientData(PortfolioError):
    """
    There have been too few portfolio transactions to perform the desired action.

    e.g. When attempting to make a withdrawal before the first deposit has been made.
    """
    pass


class InsufficientFunds(PortfolioError):
    """
    There are insufficient funds in the portfolio to perform the desired action.

    e.g. Attempting to withdraw £100 when the portfolio value is only £80.
    """
    pass



import unittest
from datetime import datetime
from unittest.mock import MagicMock

from numpy_financial import irr
from portfolio_manager.exceptions import InsufficientData
from portfolio_manager.portfolio import InvestmentPortfolio
from portfolio_manager.return_calculators import (TimeWeightedReturnCalculator,
                                                  ReturnCalculator,
                                                  SimpleReturnCalculator,
                                                  MoneyWeightedReturnCalculator)


class ReturnCalculatorsTests(unittest.TestCase):
    def setUp(self) -> None:
        self.test_portfolio = InvestmentPortfolio(name="test_portfolio")

    def test_calculate_annualised_return(self):
        # Test a portfolio age of < 1
        mock_rc = MagicMock()
        mock_rc._get_portfolio_age.return_value = 0.9
        actual = ReturnCalculator.calculate_annualised_return(mock_rc, 'test', 13)
        mock_rc._get_portfolio_age.assert_called_once_with('test')
        expected = 14.54
        self.assertEqual(expected, actual)

        # Test a portfolio age of 1
        mock_rc = MagicMock()
        mock_rc._get_portfolio_age.return_value = 1
        actual = ReturnCalculator.calculate_annualised_return(mock_rc, 'test', -11)
        mock_rc._get_portfolio_age.assert_called_once_with('test')
        expected = -11
        self.assertEqual(expected, actual)

        # Test a portfolio age of > 1
        mock_rc = MagicMock()
        mock_rc._get_portfolio_age.return_value = 2.5
        actual = ReturnCalculator.calculate_annualised_return(mock_rc, 'test', 120)
        mock_rc._get_portfolio_age.assert_called_once_with('test')
        expected = 37.08
        self.assertEqual(expected, actual)

    def test_get_portfolio_age_shorter(self):
        # Test a portfolio of age < 1 year
        test_data = [
            {
                'date': datetime(2020, 7, 1)
            },
            {
                'date': datetime(2021, 1, 1),
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual = ReturnCalculator._get_portfolio_age(self.test_portfolio)
        self.assertEqual(0.5, round(actual, 2))

    def test_get_portfolio_age_medium(self):
        # Test a portfolio of age 1 year
        test_data = [
            {
                'date': datetime(2020, 7, 1)
            },
            {
                'date': datetime(2021, 1, 1),
            },
            {
                'date': datetime(2021, 7, 1)
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual = ReturnCalculator._get_portfolio_age(self.test_portfolio)
        self.assertEqual(1, round(actual, 2))

    def test_get_portfolio_age_longer(self):
        # Test a portfolio of age > 1 year
        test_data = [
            {
                'date': datetime(2020, 7, 1)
            },
            {
                'date': datetime(2021, 1, 1),
            },
            {
                'date': datetime(2021, 1, 1),
            },
            {
                'date': datetime(2022, 9, 1)
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual = ReturnCalculator._get_portfolio_age(self.test_portfolio)
        self.assertEqual(2.17, round(actual, 2))


class SimpleReturnCalculatorTests(unittest.TestCase):
    def setUp(self) -> None:
        self.test_portfolio = InvestmentPortfolio(name="test_portfolio")
        self.simple_return_calculator = SimpleReturnCalculator()

    def test_calculate_return_no_portfolio_data(self):
        # Test that one or less transaction results in an error
        with self.assertRaises(InsufficientData):
            self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                           annualised=False)

    def test_calculate_return_negative_deposited(self):
        # This is an example of how a real portfolio could have a negative total
        # deposited value
        test_data = [
            {
                # Deposit £100
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                # Value increases to £110
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            },
            {
                # Withdraw £105, leaving £5 in the portfolio with total deposited = -£5
                'date': datetime(2021, 1, 2),
                'total_deposited': -5,
                'current_portfolio_value': 5,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        # Test that negative total deposited raises an error
        self.test_portfolio.portfolio_history = test_data
        self.test_portfolio.total_deposited = -5
        with self.assertRaises(ValueError):
            self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                           annualised=False)

    def test_calculate_return_zero_deposited(self):
        # This is an example of how a real portfolio could have a total deposited value
        # of zero
        test_data = [
            {
                # Deposit £100
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                # Value increases to £110
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            },
            {
                # Withdraw £100, leaving £10 in the portfolio with total deposited = £0
                'date': datetime(2021, 1, 2),
                'total_deposited': -5,
                'current_portfolio_value': 5,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        # Test that zero total deposited raises an error
        self.test_portfolio.portfolio_history = test_data
        self.test_portfolio.total_deposited = 0
        with self.assertRaises(ValueError):
            self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                           annualised=False)

    def test_calculate_return_positive_return(self):
        # Ensure the portfolio has non-empty history to prevent an error
        test_data = ["I'm", "not", "empty"]
        self.test_portfolio.portfolio_history = test_data

        # Test 10% return
        self.test_portfolio.total_deposited = 100
        self.test_portfolio.current_portfolio_value = 110
        actual = self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                                annualised=False)
        self.assertEqual(10, actual)

        # Test 110% return
        self.test_portfolio.total_deposited = 40
        self.test_portfolio.current_portfolio_value = 84
        actual = self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                                annualised=False)
        self.assertEqual(110, actual)

    def test_calculate_return_negative_return(self):
        # Ensure the portfolio has non-empty history to prevent an error
        test_data = ["I'm", "not", "empty"]
        self.test_portfolio.portfolio_history = test_data

        # Test -10% return
        self.test_portfolio.total_deposited = 100
        self.test_portfolio.current_portfolio_value = 90
        actual = self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                                annualised=False)
        self.assertEqual(-10, actual)

        # Test -35% return
        self.test_portfolio.total_deposited = 40
        self.test_portfolio.current_portfolio_value = 26
        actual = self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                                annualised=False)
        self.assertEqual(-35, actual)

        # Test -100% return
        self.test_portfolio.total_deposited = 40
        self.test_portfolio.current_portfolio_value = 0
        actual = self.simple_return_calculator.calculate_return(self.test_portfolio,
                                                                annualised=False)
        self.assertEqual(-100, actual)


class TimeWeightedReturnCalculatorTests(unittest.TestCase):
    def setUp(self) -> None:
        self.test_portfolio = InvestmentPortfolio(name="test_portfolio")
        self.twr_calculator = TimeWeightedReturnCalculator()

    def test_calculate_return_no_data(self):
        # Test that one or less transaction results in an error
        with self.assertRaises(InsufficientData):
            self.twr_calculator.calculate_return(self.test_portfolio, annualised=False)

    def test_calculate_return_single_period(self):
        # Test that a single sub-period reduces to the simple rate of return
        test_data = [
            {
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.twr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = 10
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_multi_deposits(self):
        # Test a multi-period with deposits only
        test_data = [
            {
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 3),
                'total_deposited': 200,
                'current_portfolio_value': 210,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 200,
                'current_portfolio_value': 215,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 5),
                'total_deposited': 250,
                'current_portfolio_value': 265,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 6),
                'total_deposited': 250,
                'current_portfolio_value': 280,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.twr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = 18.99
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_mixed_periods(self):
        # Test a multi-period including deposits and withdrawals
        test_data = [
            {
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 3),
                'total_deposited': 50,
                'current_portfolio_value': 60,
                'transaction_type': 'withdrawal'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 50,
                'current_portfolio_value': 54,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 5),
                'total_deposited': 100,
                'current_portfolio_value': 104,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 6),
                'total_deposited': 100,
                'current_portfolio_value': 119.6,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.twr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = 13.85
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_negative_return(self):
        # Test a multi-period including deposits and negative return
        test_data = [
            {
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 80,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 3),
                'total_deposited': 50,
                'current_portfolio_value': 30,
                'transaction_type': 'withdrawal'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 50,
                'current_portfolio_value': 27,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 5),
                'total_deposited': 100,
                'current_portfolio_value': 77,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 6),
                'total_deposited': 100,
                'current_portfolio_value': 78.54,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.twr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = -26.56
        self.assertEqual(actual_output, expected_output)


class MoneyWeightedReturnCalculatorTests(unittest.TestCase):
    def setUp(self) -> None:
        self.test_portfolio = InvestmentPortfolio(name="test_portfolio")
        self.mwr_calculator = MoneyWeightedReturnCalculator()

    def test_calculate_return_no_data(self):
        # Test that one or less transaction results in an error
        with self.assertRaises(ValueError):
            self.mwr_calculator.calculate_return(self.test_portfolio, annualised=False)

    def test_calculate_return_single_period(self):
        # Test that a single sub-period reduces to the simple rate of return
        test_data = [
            {
                'date': datetime(2021, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 2),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.mwr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = 10
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_deposits_and_withdrawals(self):
        # Test a multi-period with deposits and withdrawals
        test_data = [
            {
                'date': datetime(2019, 1, 1),
                'total_deposited': 50,
                'current_portfolio_value': 50,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2020, 1, 1),
                'total_deposited': 50,
                'current_portfolio_value': 52,
                'transaction_type': 'update_portfolio_value'
            },

            {
                'date': datetime(2020, 1, 1),
                'total_deposited': 48,
                'current_portfolio_value': 50,
                'transaction_type': 'withdrawal'
            },
            {
                'date': datetime(2020, 1, 3),
                'total_deposited': 48,
                'current_portfolio_value': 67,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 46,
                'current_portfolio_value': 65,
                'transaction_type': 'withdrawal'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.mwr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        # Example from: https://www.investopedia.com/terms/m/money-weighted-return.asp
        expected_output = 11.73
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_multi_deposits(self):
        # Test a multi-period portfolio with deposits only
        test_data = [
            {
                'date': datetime(2020, 1, 1),
                'total_deposited': 100,
                'current_portfolio_value': 100,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2020, 3, 1),
                'total_deposited': 100,
                'current_portfolio_value': 110,
                'transaction_type': 'update_portfolio_value'
            },

            {
                'date': datetime(2020, 6, 1),
                'total_deposited': 100,
                'current_portfolio_value': 150,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2020, 1, 3),
                'total_deposited': 200,
                'current_portfolio_value': 250,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 200,
                'current_portfolio_value': 400,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 5),
                'total_deposited': 300,
                'current_portfolio_value': 500,
                'transaction_type': 'deposit'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.mwr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = round(irr([-100, -100, -100, 500]) * 100, 2)
        self.assertEqual(actual_output, expected_output)

    def test_calculate_return_negative_return(self):
        # Test a multi-period with negative return
        test_data = [
            {
                'date': datetime(2020, 1, 1),
                'total_deposited': 50,
                'current_portfolio_value': 50,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2020, 3, 1),
                'total_deposited': 50,
                'current_portfolio_value': 30,
                'transaction_type': 'update_portfolio_value'
            },

            {
                'date': datetime(2020, 6, 1),
                'total_deposited': 100,
                'current_portfolio_value': 80,
                'transaction_type': 'deposit'
            },
            {
                'date': datetime(2020, 1, 3),
                'total_deposited': 100,
                'current_portfolio_value': 30,
                'transaction_type': 'update_portfolio_value'
            },
            {
                'date': datetime(2021, 1, 4),
                'total_deposited': 150,
                'current_portfolio_value': 80,
                'transaction_type': 'deposit'
            }
        ]
        self.test_portfolio.portfolio_history = test_data
        actual_output = self.mwr_calculator.calculate_return(self.test_portfolio,
                                                             annualised=False)
        expected_output = round(irr([-50, -50, -50, 80]) * 100, 2)
        self.assertEqual(actual_output, expected_output)


if __name__ == '__main__':
    unittest.main()



import abc
from typing import Any, Union

import numpy as np
import pandas as pd
from numpy_financial import irr

from portfolio_manager.exceptions import InsufficientData
from portfolio_manager.portfolio import InvestmentPortfolio


class ReturnCalculator(abc.ABC):
    @abc.abstractmethod
    def calculate_return(self, portfolio: InvestmentPortfolio,
                         annualised: bool) -> Any:
        pass

    def calculate_annualised_return(self, portfolio: InvestmentPortfolio,
                                    total_return_percentage: Union[int, float]):
        """
        Given the overall return percentage of a portfolio, calculate the annualised
        return percentage.
        """
        portfolio_age = self._get_portfolio_age(portfolio)
        annualised_return = (1 + total_return_percentage / 100) ** (1 / portfolio_age)
        annualised_return_percentage = (annualised_return - 1) * 100
        return round(annualised_return_percentage, 2)

    @staticmethod
    def _get_portfolio_age(portfolio: InvestmentPortfolio) -> Union[int, float]:
        """
        Calculate the 'age' of a portfolio. This is the number of years between the
        first transaction in the portfolio and the most recent one.

        Parameters
        ----------
        portfolio : InvestmentPortfolio
            The portfolio we are calculating the age of.

        Returns
        -------
        Union[int, float]: The age of the portfolio, measured in years.
        """
        # Here we make use of the fact that the transactions within portfolio history
        # are ordered by ascending date
        start = portfolio.portfolio_history[0]['date']
        end = portfolio.portfolio_history[-1]['date']
        delta = end - start

        # Return the portfolio age, in years
        return delta.days / 365.25


class SimpleReturnCalculator(ReturnCalculator):
    """
    Calculate the simple percentage return of a portfolio using the following formula:

    ((current portfolio value - total amount deposited) / total amount deposited) * 100

    Note: If, due to a combination of deposits and withdrawals, the total amount
    deposited in the portfolio is <= 0 then this value will be meaningless. In this case,
    a TimeWeightedReturnCalculator is a better choice.
    """
    def calculate_return(self, portfolio: InvestmentPortfolio,
                         annualised: bool = True) -> Any:
        """
        Calculate the simple rate of return of the portfolio.

        Parameters
        ----------
        portfolio : InvestmentPortfolio
            The portfolio we are calculating the simple return for.
        annualised : bool
            If True, calculate the annualised return.

        Returns
        -------
        Any : The simple rate of return of the portfolio, as a percentage.
            e.g. 18 represents 18%.
        """
        # If there isn't enough data in the portfolio, raise an error
        if len(portfolio.portfolio_history) <= 1:
            raise InsufficientData('Not enough portfolio data to calculate a return.')

        # If the sum of all deposits and withdrawals is negative or zero, raise an error
        if portfolio.total_deposited <= 0:
            raise ValueError('Total deposited is negative or zero.')

        return_amount = portfolio.current_portfolio_value - portfolio.total_deposited
        return_percentage = (return_amount / portfolio.total_deposited) * 100

        if annualised:
            return_percentage = self.calculate_annualised_return(portfolio,
                                                                 return_percentage)

        return round(return_percentage, 2)


class TimeWeightedReturnCalculator(ReturnCalculator):
    """
    Calculate the time-weighted rate of return of a portfolio. This metric is useful for
    measuring the performance of a portfolio where many deposits/withdrawals have been
    paid over time. See for more information
    """
    def calculate_return(self, portfolio: InvestmentPortfolio,
                         annualised: bool = True) -> Any:
        """
        Calculate the time-weighted rate of return of the portfolio.

        Parameters
        ----------
        portfolio : InvestmentPortfolio
            The portfolio we are calculating the time-weighted return for.
        annualised : bool
            If True, calculate the annualised return.

        Returns
        -------
        Any : The time-weighted rate of return of the portfolio, as a percentage.
            e.g. 18 represents 18%.
        """
        # If there isn't enough data in the portfolio, raise an error
        if len(portfolio.portfolio_history) <= 1:
            raise InsufficientData('Not enough portfolio data to calculate a return.')

        # Otherwise, calculate the time-weighted return
        sub_period_returns = []
        df = pd.DataFrame(portfolio.portfolio_history)

        # Group the data into sub-periods. Identify these sub-periods by finding groups
        # of consecutive transactions with the same 'total_deposited' values
        is_new_period_ser = pd.Series(
            df['total_deposited'].shift() != df['total_deposited'])
        grouped_periods = df.groupby(is_new_period_ser.cumsum())

        # Loop through each sub-period to calculate the returns for that period
        for _, data in grouped_periods:
            sub_period_data = pd.DataFrame(data)

            start_value = sub_period_data['current_portfolio_value'].values[0]
            end_value = sub_period_data['current_portfolio_value'].values[-1]

            return_for_period = (end_value - start_value) / start_value
            sub_period_returns.append(return_for_period + 1)

        twr_return_percentage = (np.product(sub_period_returns) - 1) * 100

        if annualised:
            twr_return_percentage = self.calculate_annualised_return(
                portfolio, twr_return_percentage)

        return round(twr_return_percentage, 2)


class MoneyWeightedReturnCalculator(ReturnCalculator):
    """
    Calculate the money-weighted rate of return of a portfolio. This metric takes into
    account the timing and volume of deposits and withdrawals.
    """

    def calculate_return(self, portfolio: InvestmentPortfolio,
                         annualised: bool = True) -> Any:
        """
        Calculate the money-weighted rate of return of the portfolio.

        Parameters
        ----------
        portfolio : InvestmentPortfolio
            The portfolio we are calculating the money-weighted return for.
        annualised : bool
            If True, calculate the annualised return.

        Returns
        -------
        Any : The money-weighted rate of return of the portfolio, as a percentage.
            e.g. 18 represents 18%.
        """
        # If there isn't enough data in the portfolio, raise an error
        if len(portfolio.portfolio_history) <= 1:
            raise ValueError('Not enough portfolio data to calculate a return.')

        # Otherwise, calculate the money-weighted return
        mwr_arr = []
        df = pd.DataFrame(portfolio.portfolio_history)

        deposits_and_withdrawals_df = df.loc[
            df['transaction_type'].isin(['deposit', 'withdrawal'])]

        # Get deposit and withdrawal amounts
        total_deposited_diff = deposits_and_withdrawals_df['total_deposited'].diff()

        mwr_arr.append(-df['total_deposited'][0])
        mwr_arr.extend(np.negative(list(total_deposited_diff)[1:]))
        mwr_arr.append(df['current_portfolio_value'].iloc[-1])

        mwr_return_percentage = (irr(mwr_arr)) * 100

        if annualised:
            mwr_return_percentage = self.calculate_annualised_return(
                portfolio, mwr_return_percentage)

        return round(mwr_return_percentage, 2)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="portfolio-manager",
    version="0.0.3",
    license="MIT",
    author="James Curwell -- George Calvert",
    author_email="jamescurwell97@gmail.com",
    description="Simple tool for monitoring the performance of investment portfolios.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/J-Curwell/portfolio-manager",
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['pandas', 'numpy', 'numpy-financial'],
    test_suite="tests"
)



import unittest
from datetime import datetime
from unittest import mock

from portfolio_manager.portfolio import InvestmentPortfolio
from portfolio_manager.exceptions import (BackDatingError, InsufficientData,
                                          InsufficientFunds)


class InvestmentPortfolioTests(unittest.TestCase):
    def setUp(self) -> None:
        self.test_portfolio = InvestmentPortfolio(name="test_portfolio")

    def test_init(self):
        self.assertIsInstance(self.test_portfolio, InvestmentPortfolio)
        self.assertEqual("test_portfolio", self.test_portfolio.name)
        self.assertEqual(0, self.test_portfolio.total_deposited)
        self.assertEqual(0, self.test_portfolio.current_portfolio_value)
        self.assertListEqual([], self.test_portfolio.portfolio_history)

    def test_deposit(self):
        # Deposit £22.50 using a specified datetime
        past = datetime(2021, 1, 1)
        self.test_portfolio.deposit(22.5, date=past)
        self.assertEqual(22.5, self.test_portfolio.total_deposited)
        self.assertEqual(22.5, self.test_portfolio.current_portfolio_value)

        # Deposit a further £27.50 using the default datetime and updating the portfolio
        # value before the deposit. To avoid time issues, we patch the value of the
        # default datetime, which is datetime.now()
        with mock.patch('portfolio_manager.portfolio.datetime') as mock_datetime:
            mock_datetime.now.return_value = datetime(2021, 2, 1)
            self.test_portfolio.deposit(27.5, portfolio_value_before_deposit=25)
        self.assertEqual(50, self.test_portfolio.total_deposited)
        self.assertEqual(52.5, self.test_portfolio.current_portfolio_value)

        # Check the two deposits and one value update were successfully recorded
        expected_transaction_history = [{'date': past,
                                         'total_deposited': 22.5,
                                         'current_portfolio_value': 22.5,
                                         'transaction_type': 'deposit'},
                                        {'date': datetime(2021, 2, 1),
                                         'total_deposited': 22.5,
                                         'current_portfolio_value': 25,
                                         'transaction_type': 'update_portfolio_value'},
                                        {'date': datetime(2021, 2, 1),
                                         'total_deposited': 50,
                                         'current_portfolio_value': 52.5,
                                         'transaction_type': 'deposit'}]
        self.assertListEqual(expected_transaction_history,
                             self.test_portfolio.portfolio_history)

        # Test that trying to incorrectly back-date a deposit raises an error
        with self.assertRaises(BackDatingError):
            self.test_portfolio.deposit(22.5, date=past)

    def test_withdraw(self):
        # Test that trying to withdraw more money than is available raises an error
        with self.assertRaises(InsufficientFunds):
            self.test_portfolio.withdraw(10)

        # Deposit some funds to avoid the above error for the remaining tests
        self.test_portfolio.deposit(50, date=datetime(2021, 1, 1))

        # Test that trying to incorrectly back-date a withdrawal raises an error
        with self.assertRaises(BackDatingError):
            self.test_portfolio.withdraw(10, date=datetime(2020, 1, 1))

        # Test withdrawing 10 from a specified date
        self.test_portfolio.withdraw(10, date=datetime(2021, 1, 2))
        self.assertEqual(40, self.test_portfolio.total_deposited)
        self.assertEqual(40, self.test_portfolio.current_portfolio_value)

        # Withdraw a further £12.50 using the default datetime and updating the portfolio
        # value before the withdrawal. To avoid time issues, we patch the value of the
        # default datetime, which is datetime.now()
        with mock.patch('portfolio_manager.portfolio.datetime') as mock_datetime:
            mock_datetime.now.return_value = datetime(2021, 2, 1)
            self.test_portfolio.withdraw(12.5, portfolio_value_before_withdrawal=25)
        self.assertEqual(27.50, self.test_portfolio.total_deposited)
        self.assertEqual(12.50, self.test_portfolio.current_portfolio_value)

        # Check the successful transactions were recorded correctly
        expected_transaction_history = [{'date': datetime(2021, 1, 1),
                                         'total_deposited': 50,
                                         'current_portfolio_value': 50,
                                         'transaction_type': 'deposit'},
                                        {'date': datetime(2021, 1, 2),
                                         'total_deposited': 40,
                                         'current_portfolio_value': 40,
                                         'transaction_type': 'withdrawal'},
                                        {'date': datetime(2021, 2, 1),
                                         'total_deposited': 40,
                                         'current_portfolio_value': 25,
                                         'transaction_type': 'update_portfolio_value'},
                                        {'date': datetime(2021, 2, 1),
                                         'total_deposited': 27.50,
                                         'current_portfolio_value': 12.5,
                                         'transaction_type': 'withdrawal'}]
        self.assertListEqual(expected_transaction_history,
                             self.test_portfolio.portfolio_history)

    def test_update_portfolio_value(self):
        # Updating the portfolio value before any deposits should cause an error
        with self.assertRaises(InsufficientData):
            self.test_portfolio.update_portfolio_value(10)

        # Deposit some funds to avoid the above error for the remaining tests
        self.test_portfolio.deposit(50, date=datetime(2021, 1, 1))

        # Test that trying to incorrectly back-date a value-update raises an error
        with self.assertRaises(BackDatingError):
            self.test_portfolio.update_portfolio_value(10, date=datetime(2020, 1, 1))

        # Test updating value to 40 from a specified date
        self.test_portfolio.update_portfolio_value(40, date=datetime(2021, 1, 2))
        self.assertEqual(50, self.test_portfolio.total_deposited)
        self.assertEqual(40, self.test_portfolio.current_portfolio_value)

        # Perform another value update using the default date. To avoid time issues, we
        # patch the value of the default datetime, which is datetime.now()
        with mock.patch('portfolio_manager.portfolio.datetime') as mock_datetime:
            mock_datetime.now.return_value = datetime(2021, 2, 1)
            self.test_portfolio.update_portfolio_value(60)
        self.assertEqual(50, self.test_portfolio.total_deposited)
        self.assertEqual(60, self.test_portfolio.current_portfolio_value)

        # Check the successful transactions were recorded correctly
        expected_transaction_history = [{'date': datetime(2021, 1, 1),
                                         'total_deposited': 50,
                                         'current_portfolio_value': 50,
                                         'transaction_type': 'deposit'},
                                        {'date': datetime(2021, 1, 2),
                                         'total_deposited': 50,
                                         'current_portfolio_value': 40,
                                         'transaction_type': 'update_portfolio_value'},
                                        {'date': datetime(2021, 2, 1),
                                         'total_deposited': 50,
                                         'current_portfolio_value': 60,
                                         'transaction_type': 'update_portfolio_value'}]
        self.assertListEqual(expected_transaction_history,
                             self.test_portfolio.portfolio_history)

    def test_update_portfolio_history(self):
        # This method has been indirectly tested above multiple times so we only add a
        # simple test here
        self.test_portfolio._update_portfolio_history(datetime(2021, 1, 1), 'testing')
        expected_transaction_history = [{'date': datetime(2021, 1, 1),
                                         'total_deposited': 0,
                                         'current_portfolio_value': 0,
                                         'transaction_type': 'testing'}]
        self.assertListEqual(expected_transaction_history,
                             self.test_portfolio.portfolio_history)

    def test_backdate_error_check(self):
        # Test that an empty portfolio doesn't raise an error
        actual = self.test_portfolio._backdate_error_check(datetime(2020, 1, 1))
        self.assertIsNone(actual)

        # Test that a transaction on the same day doesn't raise an error
        self.test_portfolio.latest_transaction_date = datetime(2020, 1, 1, 12)
        actual = self.test_portfolio._backdate_error_check(datetime(2020, 1, 1, 13))
        self.assertIsNone(actual)

        # Test that a transaction in the future doesn't raise an error
        self.test_portfolio.latest_transaction_date = datetime(2020, 1, 1)
        actual = self.test_portfolio._backdate_error_check(datetime(2020, 1, 2))
        self.assertIsNone(actual)

        # Test that a transaction in the past raises an error
        self.test_portfolio.latest_transaction_date = datetime(2020, 1, 2)
        with self.assertRaises(BackDatingError):
            self.test_portfolio._backdate_error_check(datetime(2020, 1, 1))


if __name__ == "__main__":
    unittest.main()
